diff --git a/src/thread.c b/src/thread.c
index e2deadd7a8..b1ad5e8946 100644
--- a/src/thread.c
+++ b/src/thread.c
@@ -19,6 +19,9 @@ Copyright (C) 2012-2019 Free Software Foundation, Inc.
 
 #include <config.h>
 #include <setjmp.h>
+#ifdef HAVE_GLIB
+#include <glib.h>
+#endif
 #include "lisp.h"
 #include "character.h"
 #include "buffer.h"
@@ -80,11 +83,26 @@ post_acquire_global_lock (struct thread_state *self)
 {
   struct thread_state *prev_thread = current_thread;
 
+#ifdef HAVE_GLIB
+  if (current_thread != NULL && current_thread->holding_glib_lock)
+    g_main_context_release (current_thread->context);
+#endif
+
   /* Do this early on, so that code below could signal errors (e.g.,
      unbind_for_thread_switch might) correctly, because we are already
-     running in the context of the thread pointed by SELF.  */
+     running in the context of the thread pointed to by SELF.  */
   current_thread = self;
 
+#ifdef HAVE_GLIB
+  if (current_thread->holding_glib_lock)
+    do
+      {
+	current_thread->holding_glib_lock =
+	  g_main_context_acquire (current_thread->context);
+      }
+    while (!current_thread->holding_glib_lock);
+#endif
+
   if (prev_thread != current_thread)
     {
       /* PREV_THREAD is NULL if the previously current thread
@@ -756,6 +774,10 @@ run_thread (void *state)
       }
   }
 
+#ifdef HAVE_GLIB
+  if (current_thread->holding_glib_lock)
+    g_main_context_release (current_thread->context);
+#endif
   current_thread = NULL;
   sys_cond_broadcast (&self->thread_condvar);
 
diff --git a/src/thread.h b/src/thread.h
index 498b9909c9..9e5c87000e 100644
--- a/src/thread.h
+++ b/src/thread.h
@@ -29,6 +29,10 @@ #define THREAD_H
 #include <signal.h>		/* sigset_t */
 #endif
 
+#ifdef HAVE_GLIB
+#include <glib.h>
+#endif
+
 #include "sysselect.h"		/* FIXME */
 #include "systhread.h"
 
@@ -169,6 +173,10 @@ #define getcjmp (current_thread->m_getcjmp)
      interrupter should broadcast to this condition.  */
   sys_cond_t *wait_condvar;
 
+#ifdef HAVE_GLIB
+  bool holding_glib_lock;
+  GMainContext *context;
+#endif
   /* This thread might have released the global lock.  If so, this is
      non-zero.  When a thread runs outside thread_select with this
      flag non-zero, it means it has been interrupted by SIGINT while
diff --git a/src/xgselect.c b/src/xgselect.c
index 9982a1f0e9..120b410e4d 100644
--- a/src/xgselect.c
+++ b/src/xgselect.c
@@ -54,12 +54,16 @@ xg_select (int fds_lim, fd_set *rfds, fd_set *wfds, fd_set *efds,
   GPollFD *gfds = gfds_buf;
   int gfds_size = ARRAYELTS (gfds_buf);
   int n_gfds, retval = 0, our_fds = 0, max_fds = fds_lim - 1;
-  bool context_acquired = false;
   int i, nfds, tmo_in_millisec, must_free = 0;
   bool need_to_dispatch;
 
   context = g_main_context_default ();
-  context_acquired = g_main_context_acquire (context);
+  /* Acquire the lock.  This is a busy wait for testing.  */
+  current_thread->context = context;
+  while (!current_thread->holding_glib_lock)
+    {
+      current_thread->holding_glib_lock = g_main_context_acquire (context);
+    }
   /* FIXME: If we couldn't acquire the context, we just silently proceed
      because this function handles more than just glib file descriptors.
      Note that, as implemented, this failure is completely silent: there is
@@ -70,7 +74,7 @@ xg_select (int fds_lim, fd_set *rfds, fd_set *wfds, fd_set *efds,
   if (wfds) all_wfds = *wfds;
   else FD_ZERO (&all_wfds);
 
-  n_gfds = (context_acquired
+  n_gfds = (current_thread->holding_glib_lock
 	    ? g_main_context_query (context, G_PRIORITY_LOW, &tmo_in_millisec,
 				    gfds, gfds_size)
 	    : -1);
@@ -151,7 +155,7 @@ xg_select (int fds_lim, fd_set *rfds, fd_set *wfds, fd_set *efds,
 #else
   need_to_dispatch = true;
 #endif
-  if (need_to_dispatch && context_acquired)
+  if (need_to_dispatch && current_thread->holding_glib_lock)
     {
       int pselect_errno = errno;
       /* Prevent g_main_dispatch recursion, that would occur without
@@ -164,8 +168,11 @@ xg_select (int fds_lim, fd_set *rfds, fd_set *wfds, fd_set *efds,
       errno = pselect_errno;
     }
 
-  if (context_acquired)
-    g_main_context_release (context);
+  if (current_thread->holding_glib_lock)
+    {
+      g_main_context_release (context);
+      current_thread->holding_glib_lock = false;
+    }
 
   /* To not have to recalculate timeout, return like this.  */
   if ((our_fds > 0 || (nfds == 0 && tmop == &tmo)) && (retval == 0))
